from collections import defaultdict
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

import numpy as np
import pytorch_lightning as pl
import torch
from loguru import logger
from pytorch_lightning.callbacks import Callback

from nomic import AtlasUser, atlas


class AtlasLightningContainer:
    def __init__(self):
        self.embeddings = []
        self.metadata = defaultdict(list)

    def log(
        self,
        embeddings: torch.Tensor,
        metadata: Union[
            Dict[str, List],
            Dict[str, torch.Tensor],
            Dict[str, np.ndarray],
            Dict[str, int],
            Dict[str, float],
            Dict[str, str],
        ] = {},
    ):
        """Log a batch of embeddings and corresponding metadata for each embedding."""
        assert isinstance(embeddings, torch.Tensor), "You must log a torch Tensor"

        # ensure passed in embeddings are a tensor with shape (batch_size, dimension)
        if len(embeddings.shape) != 2:
            raise ValueError("Your logged embedding tensor must have shape (N,d)")

        metadata_copy = defaultdict(list)
        # sanity check the inputs
        for key in metadata:
            metadata_value = metadata[key]
            if isinstance(metadata_value, torch.Tensor):
                metadata_copy[key] = metadata_value.flatten().cpu().tolist()
            elif isinstance(metadata_value, np.ndarray):
                metadata_copy[key] = metadata_value.flatten().tolist()
            else:
                if not isinstance(metadata_value, list):
                    if (
                        isinstance(metadata_value, float)
                        or isinstance(metadata_value, int)
                        or isinstance(metadata_value, str)
                    ):
                        metadata_copy[key] = [metadata_value]

            if embeddings.shape[0] != len(metadata_copy[key]):
                raise ValueError(
                    f"Your metadata has invalid shape. You have {embeddings.shape[0]} embeddings but len(metadata['{key}']) = {len(metadata_copy[key])}"
                )

        self.embeddings.append(embeddings)
        for key in metadata:
            self.metadata[key] = self.metadata[key] + metadata_copy[key]

    def clear(self):
        """Clears all embeddings and metadata from the final produced map."""
        self.embeddings = []
        self.metadata = defaultdict(list)


class AtlasLightningModule(pl.LightningModule):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self.atlas: Optional[AtlasLightningContainer] = None


class AtlasEmbeddingExplorer(Callback):
    def __init__(
        self,
        max_points=-1,
        rebuild_time_delay=600,
        name=None,
        description="",
        is_public=True,
        overwrite_on_validation=False,
    ):
        """

        Args:
            max_points: The maximum points to visualize. -1 will visualize all points.
            rebuild_time_delay: Only rebuilds the embedding explorer if 'rebuild_time_delay' seconds have passed since the last rebuild.
            name: The name for your embedding explorer.
            description: A description for your embedding explorer.
            is_public: Should your embedding explorer be public
            overwrite_on_validation: Re-creates your validation set viewer on every validation run.
        """
        self.max_points = max_points
        self.name = name
        self.description = description
        self.is_public = is_public
        self.overwrite = overwrite_on_validation
        self.dataset = None
        self.map = None
        self.atlas = AtlasLightningContainer()
        self.rebuild_time_delay = rebuild_time_delay
        self.last_rebuild_timestamp = datetime.min

    def on_train_start(self, trainer: "pl.Trainer", pl_module: pl.LightningModule):
        """Verify that atlas is configured and set up variables"""
        AtlasUser()  # verify logged in.
        AtlasLightningModule(pl_module).atlas = self.atlas

    def on_sanity_check_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        AtlasUser()  # verify logged in.
        AtlasLightningModule(pl_module).atlas = AtlasLightningContainer()

    def on_train_epoch_start(self, *args, **kwargs):
        self.atlas.clear()

    def on_validation_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        self.atlas.clear()

    def on_train_end(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        pass

    def on_validation_end(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        self._create_map()

    def _create_map(self):
        seconds_since_last_build = int((datetime.now() - self.last_rebuild_timestamp).total_seconds())
        if seconds_since_last_build < self.rebuild_time_delay:
            logger.info(
                f"Skipping regenerating embedding explorer for this validation epoch as its been {seconds_since_last_build} seconds since the last rebuild and `rebuild_time_delay={self.rebuild_time_delay}`"
            )
            if self.map is not None:
                logger.info(f" See your previous validation embedding space at: {self.map.map_link}")
            return
        if not self.atlas.embeddings:
            logger.info(
                "Pytorch module does not have logits recorded in _atlas_logits property, AtlasEmbeddingExplorer will not generate a map."
            )
            return
        embeddings = torch.cat(self.atlas.embeddings).detach().cpu().numpy()

        lengths = map(len, self.atlas.metadata.values())
        assert set(lengths) != 1, "Error in logging metadata, it is not all the same length"

        if self.max_points > 0:
            embeddings = embeddings[: self.max_points, :]
            for key in self.atlas.metadata:
                self.atlas.metadata[key] = self.atlas.metadata[key][: self.max_points]

                # convert all uniqish values to strings prior to insertion into Atlas so
                if len(set(self.atlas.metadata[key])) <= 10:
                    self.atlas.metadata[key] = [str(x) for x in self.atlas.metadata[key]]

        keys = list(self.atlas.metadata.keys())
        if "id" not in self.atlas.metadata:
            self.atlas.metadata["id"] = [i for i in range(embeddings.shape[0])]
            keys.append("id")

        metadata = [dict(zip(keys, vals)) for vals in zip(*(self.atlas.metadata[k] for k in keys))]

        colorable_fields = []
        for key in keys:
            if key == "id":
                continue
            if (
                isinstance(metadata[0][key], float)
                or isinstance(metadata[0][key], int)
                or key in ("class", "label", "target")
                or len(set(self.atlas.metadata[key])) <= 10
            ):
                colorable_fields.append(key)

        try:
            project = atlas.map_data(
                embeddings=embeddings,
                data=metadata,
                id_field="id",
                is_public=self.is_public,
                identifier=self.name,
                description=self.description,
                topic_model=False,
            )
        except BaseException as e:
            logger.info(e)
            logger.info("Failed to update your map on this validation epoch.")
            return
        self.dataset = project
        self.map = project.maps[0]
        self.last_rebuild_timestamp = datetime.now()



"""
This class allows for programmatic interactions with Atlas - Nomic's neural database. Initialize AtlasClient in any Python context such as a script
or in a Jupyter Notebook to organize and interact with your unstructured data.
"""

import uuid
from typing import Dict, Iterable, List, Optional, Union

import numpy as np
import pyarrow as pa
from loguru import logger
from pandas import DataFrame
from PIL import Image
from pyarrow import Table
from tqdm import tqdm

from .data_inference import NomicDuplicatesOptions, NomicEmbedOptions, NomicProjectOptions, NomicTopicOptions
from .dataset import AtlasDataset, AtlasDataStream
from .settings import *
from .utils import arrow_iterator, b64int, get_random_name


def map_data(
    data: Optional[Union[DataFrame, List[Dict], Table]] = None,
    blobs: Optional[List[Union[str, bytes, Image.Image]]] = None,
    embeddings: Optional[np.ndarray] = None,
    identifier: Optional[str] = None,
    description: str = "",
    id_field: Optional[str] = None,
    is_public: bool = True,
    indexed_field: Optional[str] = None,
    projection: Union[bool, Dict, NomicProjectOptions] = True,
    topic_model: Union[bool, Dict, NomicTopicOptions] = True,
    duplicate_detection: Union[bool, Dict, NomicDuplicatesOptions] = True,
    embedding_model: Optional[Union[str, Dict, NomicEmbedOptions]] = None,
) -> AtlasDataset:
    """

    Args:
        data: An ordered collection of the datapoints you are structuring. Can be a list of dictionaries, Pandas Dataframe or PyArrow Table.
        blobs: A list of image paths, bytes, or PIL images to add to your image dataset that are stored locally.
        embeddings: An [N,d] numpy array containing the N embeddings to add.
        identifier: A name for your dataset that is used to generate the dataset identifier. A unique name will be chosen if not supplied.
        description: The description of your dataset
        id_field: Specify your data unique id field. This field can be up 36 characters in length. If not specified, one will be created for you named `id_`.
        is_public: Should the dataset be accessible outside your Nomic Atlas organization.
        indexed_field: The text field from the dataset that will be used to create embeddings, which determines the layout of the data map in Atlas. Required for text data but won't have an impact if uploading embeddings or image blobs.
        projection: Options to adjust Nomic Project - the dimensionality algorithm organizing your dataset.
        topic_model: Options to adjust Nomic Topic - the topic model organizing your dataset.
        duplicate_detection: Options to adjust Nomic Duplicates - the duplicate detection algorithm.
        embedding_model: Options to adjust the embedding model used to embed your dataset.
    :return:
    """
    modality = "embedding"
    if embeddings is not None:
        assert isinstance(embeddings, np.ndarray), "You must pass in a numpy array"
        if embeddings.size == 0:
            raise Exception("Your embeddings cannot be empty")

    if indexed_field is not None:
        if embeddings is not None:
            logger.warning("You have specified an indexed field but are using embeddings. Embeddings will be ignored.")
        modality = "text"

    if blobs is not None:
        if embeddings is not None:
            raise ValueError(
                "You cannot pass both `blobs` and `embeddings` to map_data(). To create a map of images, include `blobs` and not `embeddings`. To create a map of embeddings with images as metadata, include your images as a field in your `data` parameter."
            )
        # change this when we support other modalities
        modality = "image"
        indexed_field = "_blob_hash"
        if embedding_model is not None:
            if isinstance(embedding_model, str):
                model_name = embedding_model
            elif isinstance(embedding_model, dict):
                model_name = embedding_model["model"]
            elif isinstance(embedding_model, NomicEmbedOptions):
                model_name = embedding_model.model
            else:
                raise ValueError("embedding_model must be a string, dictionary, or NomicEmbedOptions object")

            if model_name in ["nomic-embed-text-v1", "nomic-embed-text-v1.5"]:
                raise Exception("You cannot use a text embedding model with blobs")
        else:
            # default to vision v1.5
            embedding_model = NomicEmbedOptions(model="nomic-embed-vision-v1.5")

    if id_field is None:
        id_field = ATLAS_DEFAULT_ID_FIELD

    project_name = get_random_name()

    dataset_name = project_name
    index_name = dataset_name

    if identifier:
        dataset_name = identifier
        index_name = identifier
    if description:
        description = description

    # no metadata was specified
    added_id_field = False

    if data is None:
        added_id_field = True
        if embeddings is not None:
            data = [{ATLAS_DEFAULT_ID_FIELD: b64int(i)} for i in range(len(embeddings))]
        elif blobs is not None:
            data = [{ATLAS_DEFAULT_ID_FIELD: b64int(i)} for i in range(len(blobs))]
        else:
            raise ValueError("You must specify either data, embeddings, or blobs")

    if id_field == ATLAS_DEFAULT_ID_FIELD and data is not None:
        if isinstance(data, list) and id_field not in data[0]:
            added_id_field = True
            for i in range(len(data)):
                # do not modify object the user passed in - also ensures IDs are unique if two input datums are the same *object*
                data[i] = data[i].copy()
                data[i][id_field] = b64int(i)
        elif isinstance(data, DataFrame) and id_field not in data.columns:
            data[id_field] = [b64int(i) for i in range(data.shape[0])]
            added_id_field = True
        elif isinstance(data, pa.Table) and not id_field in data.column_names:  # type: ignore
            ids = pa.array([b64int(i) for i in range(len(data))])
            data = data.append_column(id_field, ids)  # type: ignore
            added_id_field = True
        elif id_field not in data[0]:
            raise ValueError("map_data data must be a list of dicts, a pandas dataframe, or a pyarrow table")

    if added_id_field:
        logger.warning("An ID field was not specified in your data so one was generated for you in insertion order.")

    dataset = AtlasDataset(
        identifier=dataset_name, description=description, unique_id_field=id_field, is_public=is_public
    )

    number_of_datums_before_upload = dataset.total_datums

    if number_of_datums_before_upload > 0:
        raise Exception("Cannot use map_data to update an existing dataset.")

    # Add data by modality
    logger.info("Uploading data to Atlas.")
    try:
        if modality == "text":
            dataset.add_data(data=data)
        elif modality == "embedding":
            dataset.add_data(
                embeddings=embeddings,
                data=data,
            )
        elif modality == "image":
            dataset.add_data(blobs=blobs, data=data)

    except BaseException as e:
        if number_of_datums_before_upload == 0:
            logger.info(f"{dataset.identifier}: Deleting dataset due to failure in initial upload.")
            dataset.delete()
        raise e

    logger.info(f"`{dataset.identifier}`: Data upload succeeded to dataset`")

    dataset.create_index(
        name=index_name,
        indexed_field=indexed_field,
        modality=modality,
        projection=projection,
        topic_model=topic_model,
        duplicate_detection=duplicate_detection,
        embedding_model=embedding_model,
    )

    dataset = dataset._latest_dataset_state()
    return dataset


def map_embeddings(
    embeddings: np.ndarray,
    data: Optional[List[Dict]] = None,
    id_field: Optional[str] = None,
    name: Optional[str] = None,
    description: Optional[str] = None,
    is_public: bool = True,
    colorable_fields: list = [],
    build_topic_model: bool = True,
    topic_label_field: Optional[str] = None,
    num_workers: None = None,
    reset_project_if_exists: bool = False,
    add_datums_if_exists: bool = False,
    shard_size: None = None,
    projection_n_neighbors: int = DEFAULT_PROJECTION_N_NEIGHBORS,
    projection_epochs: int = DEFAULT_PROJECTION_EPOCHS,
    projection_spread: float = DEFAULT_PROJECTION_SPREAD,
    organization_name=None,
) -> AtlasDataset:
    """

    Args:
        embeddings: An [N,d] numpy array containing the batch of N embeddings to add.
        data: An [N,] element list of dictionaries containing metadata for each embedding.
        id_field: Specify your data unique id field. This field can be up 36 characters in length. If not specified, one will be created for you named `id_`.
        name: A name for your dataset. Specify in the format `organization/project` to create in a specific organization.
        description: A description for your map.
        is_public: Should this embedding map be public? Private maps can only be accessed by members of your organization.
        colorable_fields: The dataset fields you want to be able to color by on the map. Must be a subset of the projects fields.
        reset_project_if_exists: If the specified dataset exists in your organization, reset it by deleting all of its data. This means your uploaded data will not be contextualized with existing data.
        add_datums_if_exists: If specifying an existing dataset and you want to add data to it, set this to true.
        build_topic_model: Builds a hierarchical topic model over your data to discover patterns.
        topic_label_field: The metadata field to estimate topic labels from. Usually the field you embedded.
        projection_n_neighbors: The number of neighbors to build.
        projection_epochs: The number of epochs to build the map with.
        projection_spread: The spread of the map.

    Returns:
        An AtlasDataset that now contains your map.

    """

    assert isinstance(embeddings, np.ndarray), "You must pass in a numpy array"
    raise DeprecationWarning("map_embeddings is deprecated and will soon be removed, use atlas.map_data instead.")


def map_text(
    data: Union[Iterable[Dict], DataFrame],
    indexed_field: str,
    id_field: Optional[str] = None,
    name: Optional[str] = None,
    description: Optional[str] = None,
    build_topic_model: bool = True,
    is_public: bool = True,
    colorable_fields: list = [],
    num_workers: None = None,
    reset_project_if_exists: bool = False,
    add_datums_if_exists: bool = False,
    shard_size: None = None,
    projection_n_neighbors: int = DEFAULT_PROJECTION_N_NEIGHBORS,
    projection_epochs: int = DEFAULT_PROJECTION_EPOCHS,
    projection_spread: float = DEFAULT_PROJECTION_SPREAD,
    duplicate_detection: bool = True,
    duplicate_threshold: float = DEFAULT_DUPLICATE_THRESHOLD,
    organization_name=None,
) -> AtlasDataset:
    """
    Generates or updates a map of the given text.

    Args:
        data: An [N,] element iterable of dictionaries containing metadata for each embedding.
        indexed_field: The name the data field containing the text your want to map.
        id_field: Specify your data unique id field. This field can be up 36 characters in length. If not specified, one will be created for you named `id_`.
        name: A name for your dataset. Specify in the format `organization/project` to create in a specific organization.
        description: A description for your map.
        build_topic_model: Builds a hierarchical topic model over your data to discover patterns.
        is_public: Should this embedding map be public? Private maps can only be accessed by members of your organization.
        colorable_fields: The dataset fields you want to be able to color by on the map. Must be a subset of the projects fields.
        reset_project_if_exists: If the specified dataset exists in your organization, reset it by deleting all of its data. This means your uploaded data will not be contextualized with existing data.
        add_datums_if_exists: If specifying an existing dataset and you want to add data to it, set this to true.
        projection_n_neighbors: The number of neighbors to build.
        projection_epochs: The number of epochs to build the map with.
        projection_spread: The spread of the map.

    Returns:
        The AtlasDataset containing your map.

    """
    if organization_name is not None:
        logger.warning(
            "Passing organization name has been removed in Nomic Python client 3.0. Instead identify your dataset with `organization_name/project_name` (e.g. sterling-cooper/november-ads)."
        )
    raise DeprecationWarning("map_text is deprecated and will soon be removed, use atlas.map_data instead.")


# NOTE: This will be deprecated for AtlasDataStream class
def _get_datastream_credentials(name: Optional[str] = "contrastors") -> Dict[str, str]:
    """
    Returns credentials for a datastream.

    Args:
        name: Datastream name
    Returns:
        A dictionary with credentials to access a datastream.
    """
    atlas_data_access = AtlasDataStream(name)
    return atlas_data_access.get_credentials()



from typing import Any, Dict, List, Literal, Optional, Union

import pyarrow as pa
from pydantic import AliasChoices, BaseModel, Field

from .settings import DEFAULT_DUPLICATE_THRESHOLD


def from_list(values: Dict[str, Any], schema=None) -> pa.Table:
    tb = pa.Table.from_pylist(values, schema=schema)
    return tb


permitted_types = {
    "integer": pa.int32(),
    "float": pa.float32(),
    "date": pa.timestamp("ms"),
    "string": pa.string(),
    "categorical": pa.string(),
}


def convert_pyarrow_schema_for_atlas(schema: pa.Schema) -> pa.Schema:
    """
    Convert a pyarrow schema to one with types that match the subset of types supported by Atlas for upload.
    """
    types = {}
    whitelist = {}
    for field in schema:
        if field.name.startswith("_"):
            # Underscore fields are private to Atlas and will be handled with their own logic.
            if not field.name in {"_embeddings", "_blob_hash"}:
                raise ValueError(f"Underscore fields are reserved for Atlas internal use: {field.name}")
            whitelist[field.name] = field.type
        elif pa.types.is_boolean(field.type):
            raise TypeError(f"Boolean type not supported: {field.name}")
        elif pa.types.is_list(field.type):
            raise TypeError(f"List types not supported: {field.name}")
        elif pa.types.is_struct(field.type):
            raise TypeError(f"Struct types not supported: {field.name}")
        elif pa.types.is_dictionary(field.type):
            types[field.name] = "categorical"
        elif pa.types.is_string(field.type):
            types[field.name] = "string"
        elif pa.types.is_integer(field.type):
            types[field.name] = "integer"
        elif pa.types.is_floating(field.type):
            types[field.name] = "float"
        elif pa.types.is_timestamp(field.type):
            types[field.name] = "date"
        elif pa.types.is_temporal(field.type):
            types[field.name] = "date"
        else:
            raise TypeError(f"Unknown type: {field.name} {field.type}")
    usertypes = {k: permitted_types[v] for k, v in types.items()}

    return pa.schema({**usertypes, **whitelist})


class NomicProjectOptions(BaseModel):
    """
    Options for Nomic 2D Dimensionality Reduction Model

    Args:
        n_neighbors: The number of neighbors to use when approximating the high dimensional embedding space during reduction. Default: `None` (Auto-inferred).
        n_epochs: How many dataset passes to train the projection model. Default: `None` (Auto-inferred).
        model: The model to use when generating the 2D projected embedding space layout. Possible values: `None` or `nomic-project-v1` or `nomic-project-v2`. Default: `None`.
        local_neighborhood_size: Only used when model is `nomic-project-v2`. Controls the size of the neighborhood used in the local structure optimizing step of `nomic-project-v2` algorithm. Min value: `max(n_neighbors, 1)`; max value: `128`. Default: `None` (Auto-inferred).
        spread: Determines how tight together points appear. Larger values result a more spread out point layout. Min value: `0`. It is recommended leaving this value as the default `None` (Auto-inferred).
        rho: Only used when model is nomic-project-v2. Controls the spread in the local structure optimizing step of `nomic-project-v2`. Min value: `0`; max value: `1`. It is recommended to leave this value as the default `None` (Auto-inferred).
    """

    n_neighbors: Optional[int] = None
    n_epochs: Optional[int] = None
    spread: Optional[float] = None
    local_neighborhood_size: Optional[int] = None
    model: Optional[str] = None
    rho: Optional[float] = None


class NomicTopicOptions(BaseModel):
    """
    Options for Nomic Topic Model

    Args:
        build_topic_model: If True, builds a topic model over your dataset's embeddings.
        topic_label_field: The dataset column (usually the column you embedded) that Atlas will use to assign a human-readable description to each topic.
    """

    build_topic_model: bool = True
    topic_label_field: Optional[str] = Field(default=None)
    cluster_method: str = "fast"
    enforce_topic_hierarchy: bool = False


class NomicDuplicatesOptions(BaseModel):
    """
    Options for Duplicate Detection

    Args:
        tag_duplicates: Should duplicate detection run over your datasets embeddings?
        duplicate_cutoff: A hyperparameter of duplicate detection, smaller values capture more exact duplicates.
    """

    tag_duplicates: bool = True
    duplicate_cutoff: float = DEFAULT_DUPLICATE_THRESHOLD


class NomicEmbedOptions(BaseModel):
    """
    Options for Configuring the Nomic Embedding Model

    Args:
        model: The Nomic Embedding Model to use.
    """

    model: Literal[
        "nomic-embed-text-v1",
        "nomic-embed-vision-v1",
        "nomic-embed-text-v1.5",
        "nomic-embed-vision-v1.5",
        "gte-multilingual-base",
    ] = "nomic-embed-text-v1.5"



import concurrent.futures
import io
import json
import logging
import multiprocessing as mp
from pathlib import PosixPath
from typing import List, Optional, Tuple, Union

import boto3
import PIL
import PIL.Image
import sagemaker
from tqdm import tqdm

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def _get_sagemaker_role():
    try:
        return sagemaker.get_execution_role()
    except ValueError:
        raise ValueError("Unable to fetch sagemaker execution role. Please provide a role.")


def parse_sagemaker_response(response):
    """
    Parse response from a sagemaker server and return the embedding as a numpy array.

    Args:
        response: The response from the sagemaker server.

    Returns:
        np.float16 array of embeddings.
    """
    # Parse json header size length from the response
    resp = json.loads(response["Body"].read().decode())
    return resp["embeddings"]


def batch_transform_text(
    s3_input_path: str,
    s3_output_path: str,
    region_name: str,
    arn: Optional[str] = None,
    role: Optional[str] = None,
    max_payload: Optional[int] = 6,
    instance_type: str = "ml.p3.2xlarge",
    n_instances: int = 1,
    wait: bool = True,
    logs: bool = True,
):
    """
    Batch transform a list of texts using a sagemaker model.

    Args:
        s3_input_path: S3 path to the input data. Input data should be a csv file without any column headers
            with each line containing a single text.
        s3_output_path: S3 path to save the output embeddings. Embeddings will be in order of the input data.
        region_name: AWS region to use.
        arn: The model package arn to use.
        role: Arn of the IAM role to use (must have permissions to S3 data as well).
        max_payload: The maximum payload size in megabytes.
        instance_type: The instance type to use.
        n_instances: The number of instances to use.
        wait: Whether method should wait for job to finish.
        logs: Whether to log the job status (only meaningful if wait is True).
    Returns:
        The job name.
    """
    if arn is None:
        raise ValueError("model package arn is currently required.")

    if role is None:
        logger.info("No role provided. Using default sagemaker role.")
        role = _get_sagemaker_role()

    sm_client = boto3.client("sagemaker", region_name=region_name)
    sm_session = sagemaker.Session(
        boto_session=boto3.Session(region_name=region_name),
        sagemaker_client=sm_client,
    )
    model = sagemaker.ModelPackage(
        name=arn.split("/")[-1],
        role=role,
        model_data=None,
        sagemaker_session=sm_session,  # makes sure the right region is used
        model_package_arn=arn,
    )
    embedder = model.transformer(
        instance_count=n_instances,
        instance_type=instance_type,
        output_path=s3_output_path,
        strategy="MultiRecord",
        assemble_with="Line",
        max_payload=max_payload,
    )
    embedder.transform(
        data=s3_input_path,
        content_type="text/csv",
        split_type="Line",
        wait=wait,
        logs=logs,
    )
    job_name = None
    if embedder.latest_transform_job is not None:
        job_name = embedder.latest_transform_job.name
    return job_name


def embed_text(
    texts: List[str],
    sagemaker_endpoint: str,
    region_name: str,
    task_type: str = "search_document",
    batch_size: int = 32,
    dimensionality: int = 768,
    binary: bool = False,
):
    """
    Embed a list of texts using a sagemaker model endpoint.

    Args:
        texts: List of texts to be embedded.
        sagemaker_endpoint: The sagemaker endpoint to use.
        region_name: AWS region sagemaker endpoint is in.
        task_type: The task type to use when embedding.
        batch_size: Size of each batch. Default is 32.
        dimensionality: Number of dimensions to return. Options are (64, 128, 256, 512, 768).
        binary: Whether to return binary embeddings.

    Returns:
        Dictionary with "embeddings" (python 2d list of floats), "model" (sagemaker endpoint used to generate embeddings).
    """

    if len(texts) == 0:
        logger.warning("No texts to embed.")
        return None

    assert task_type in [
        "search_query",
        "search_document",
        "classification",
        "clustering",
    ], f"Invalid task type: {task_type}"

    assert dimensionality in (
        64,
        128,
        256,
        512,
        768,
    ), f"Invalid number of dimensions: {dimensionality}"

    client = boto3.client("sagemaker-runtime", region_name=region_name)
    embeddings = []

    for i in tqdm(range(0, len(texts), batch_size)):
        batch = json.dumps(
            {
                "texts": texts[i : i + batch_size],
                "binary": binary,
                "dimensionality": dimensionality,
                "task_type": task_type,
            }
        )
        response = client.invoke_endpoint(EndpointName=sagemaker_endpoint, Body=batch, ContentType="application/json")
        embeddings.extend(parse_sagemaker_response(response))

    return {
        "embeddings": embeddings,
        "model": "nomic-embed-text-v1.5",
        "usage": {},
    }


# only way I could get sagemaker with multipart to work
def prepare_multipart_request(images: List[Tuple[str, bytes]]) -> Tuple[bytes, bytes]:
    # Prepare the multipart body
    boundary = b"---------------------------Boundary"
    body = b""
    for i, (name, img_bytes) in enumerate(images):
        body += b"--" + boundary + b"\r\n"
        body += f'Content-Disposition: form-data; name="{name}"; filename="image_{i}.jpg"\r\n'.encode("utf-8")
        body += b"Content-Type: image/jpeg\r\n\r\n"
        body += img_bytes + b"\r\n"
    body += b"--" + boundary + b"--\r\n"

    return body, boundary


def preprocess_image(images: List[Union[str, "PIL.Image.Image", bytes]]) -> Tuple[bytes, bytes]:
    """
    Preprocess a list of images for embedding using a sagemaker model.

    Args:
        images: List of images to be embedded.

    Returns:
        List of images formatted for sagemaker embedding.
    """
    encoded_images = []
    for image in images:
        if isinstance(image, str):
            image = PIL.Image.open(image)
        elif isinstance(image, bytes):
            image = PIL.Image.open(io.BytesIO(image))
        elif isinstance(image, PosixPath):
            image = PIL.Image.open(image)
        else:
            assert isinstance(image, PIL.Image.Image), f"Invalid image type: {type(image)}"
        image = image.convert("RGB")
        buffered = io.BytesIO()
        image.save(buffered, format="JPEG")
        encoded_images.append(("image_data", buffered.getvalue()))

    body, boundary = prepare_multipart_request(encoded_images)
    return body, boundary


def sagemaker_image_request(
    images: List[Union[str, bytes, "PIL.Image.Image"]], sagemaker_endpoint: str, region_name: str
):
    body, boundary = preprocess_image(images)

    client = boto3.client("sagemaker-runtime", region_name=region_name)
    response = client.invoke_endpoint(
        EndpointName=sagemaker_endpoint,
        Body=body,
        ContentType=f'multipart/form-data; boundary={boundary.decode("utf-8")}',
    )

    return parse_sagemaker_response(response)


def embed_image(
    images: List[Union[str, "PIL.Image.Image", bytes]],
    sagemaker_endpoint: str,
    region_name: str,
    model_name="nomic-embed-vision-v1.5",
    batch_size=16,
) -> dict:
    embeddings = []

    pbar = tqdm(total=len(images))
    for i in range(0, len(images), batch_size):
        batch = images[i : i + batch_size]
        embeddings.extend(
            sagemaker_image_request(batch, sagemaker_endpoint=sagemaker_endpoint, region_name=region_name)
        )
        pbar.update(len(batch))

    return {
        "embeddings": embeddings,
        "model": model_name,
        "usage": {},
    }


def batch_transform_image(
    s3_input_path: str,
    s3_output_path: str,
    region_name: str,
    arn: Optional[str] = None,
    role: Optional[str] = None,
    max_payload: Optional[int] = 6,
    instance_type: str = "ml.g4dn.xlarge",
    n_instances: int = 1,
    wait: bool = True,
    logs: bool = True,
):
    """
    Batch transform a list of texts using a sagemaker model.

    Args:
        s3_input_path: S3 path to the input data. Input data should be a csv file without any column headers
            with each line containing a single text.
        s3_output_path: S3 path to save the output embeddings. Embeddings will be in order of the input data.
        region_name: AWS region to use.
        arn: The model package arn to use.
        role: Arn of the IAM role to use (must have permissions to S3 data as well).
        max_payload: The maximum payload size in megabytes.
        instance_type: The instance type to use.
        n_instances: The number of instances to use.
        wait: Whether method should wait for job to finish.
        logs: Whether to log the job status (only meaningful if wait is True).
    Returns:
        The job name.
    """
    if arn is None:
        raise ValueError("model package arn is currently required.")

    if role is None:
        logger.info("No role provided. Using default sagemaker role.")
        role = _get_sagemaker_role()

    sm_client = boto3.client("sagemaker", region_name=region_name)
    sm_session = sagemaker.Session(
        boto_session=boto3.Session(region_name=region_name),
        sagemaker_client=sm_client,
    )
    model = sagemaker.ModelPackage(
        name=arn.split("/")[-1],
        role=role,
        model_data=None,
        sagemaker_session=sm_session,  # makes sure the right region is used
        model_package_arn=arn,
    )
    embedder = model.transformer(
        instance_count=n_instances,
        instance_type=instance_type,
        output_path=s3_output_path,
        assemble_with="Line",
        max_payload=max_payload,
    )
    embedder.transform(
        data=s3_input_path,
        content_type="application/x-image",
        split_type=None,
        wait=wait,
        logs=logs,
    )
    job_name = None
    if embedder.latest_transform_job is not None:
        job_name = embedder.latest_transform_job.name
    return job_name
